"""
Exports {{ variable_name }} table keyed by supporting unicode version level.

This code generated by python wcwidth project.
"""
# pylint: disable=duplicate-code
{{ variable_name }} = {
{%- for table_version, table_def in table.items() %}
    '{{ table_version }}': (
        # Source: {{ table_def.filename }}
        # Date: {{ table_def.date }}
        #
{%- for hex_start, hex_end, txt_description in table_def.hex_range_descriptions %}
        ({{ hex_start }}, {{ hex_end }},),  # {{txt_description}}
{%- endfor %}
    ),
{%- endfor %}
}
