"""
Grapheme overrides for {{ canonical_name }}.

This code generated by python wcwidth using ucs-detect project data.

Terminals: {{ terminals|join(', ') }}
"""
# pylint: skip-file
GRAPHEMES = {
{%- for grapheme_str, terminal_width in graphemes | dictsort %}
    '{{ grapheme_str }}': {{ terminal_width }},
{%- endfor %}
}
