"""
Terminal to grapheme mapping of overrides, using a hash registry.

This code generated by python wcwidth using ucs-detect project data.
"""
# pylint: skip-file
_REGISTRY = {
{%- for terminal, hash_key in registry | dictsort %}
    {{ terminal.__repr__() }}: {{ hash_key.__repr__() }},
{%- endfor %}
}

__all__ = ('_REGISTRY',)
