"""
Exports {{ variable_name }} table keyed by canonical terminal software name.

Contains multi-codepoint grapheme overrides that cannot be expressed as
codepoint ranges.

This code generated by python wcwidth using ucs-detect project data.
"""
# pylint: disable=duplicate-code
{{ variable_name }} = {
{%- for term_name, graphemes in table.items() %}
    '{{ term_name }}': {
{%- for grapheme_str, terminal_width in graphemes.items() %}
        '{{ grapheme_str }}': {{ terminal_width }},
{%- endfor %}
    },
{%- endfor %}
}
