"""
Exports grapheme cluster break property tables for Unicode version {{ unicode_version }}.

This module provides lookup tables for Unicode grapheme cluster break properties as defined in UAX
#29: Unicode Text Segmentation.

This code generated by python wcwidth project.
"""
# pylint: disable=duplicate-code
{%- for var_name, table_def in tables.items() %}

{{ var_name }} = (
    # Source: {{ table_def.filename }}
    # Date: {{ table_def.date }}
    #
{%- for hex_start, hex_end, txt_description in table_def.hex_range_descriptions %}
    ({{ hex_start }}, {{ hex_end }},),  # {{txt_description}}
{%- endfor %}
)
{%- endfor %}
