=====================
Unicode release files
=====================

This library aims to be forward-looking, portable, and most correct.
The most current release of this API is based on the Unicode Standard
release files:

{% for line1, line2 in source_headers %}
``{{line1}}``
  *{{line2}}*
{% endfor %}
